/*
 * Nintendo 3DS Ticket Parser
 * Copyright (c) 2017 Eitan Levy. Public Domain.
 *
 * The program parse and print structure of the specified ticket dump,
 * including certificate information, etc.

 * Compile: cc -O2 TicketParser.c -o TicketParser
 * Run: ./TicketParser <path-of-ticket>
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <stdint.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

// Valid Signature Types
enum { 
    R_RSA_4096_SHA1     = 0x010000, 
    R_RSA_2048_SHA1     = 0x010001, 
    R_EC_SHA1           = 0x010002,
    R_RSA_4096_SHA256   = 0x010003,
    R_RSA_2048_SHA256   = 0x010004,
    R_ECDSA_SHA256      = 0x010005
} SignatureType;

typedef struct { const char *desc; uint32_t szData, szPadding; } signInfoTab; 
static signInfoTab signInfo[6] = {
    { "RSA 4096 with SHA1 (UNUSED)",     0x200,  0x3c },
    { "RSA 2048 with SHA1 (UNUSED)",     0x100,  0x3c },
    { "ECC with SHA1 (UNUSED)",          0x3c,   0x40 },
    { "RSA 4096 with SHA256",            0x200,  0x3c },
    { "RSA 2048 with SHA256",            0x100,  0x3c },
    { "ECDSA with SHA256",               0x3c,   0x40 }
};
#define getSigInfoIdx(t) t - 0x010000
#define getSignDataLength(t) signInfo[getSigInfoIdx(t)].szData
#define getSignPaddingSize(t) signInfo[getSigInfoIdx(t)].szPadding
#define getSignDesc(t) signInfo[getSigInfoIdx(t)].desc

typedef struct {
    uint32_t signType;
    void *   signData;
} SignatureData;

typedef struct {
    uint8_t     issuer[0x40],           // Ticket Issuer
                pubkey[0x3c],           // ECC Public Key
                version,                // Ticket Version. Expected 0x1 on 3DS
                caCrlVersion,           // CaCrlVersion
                signerCrlVersion;       // SignerCrlVersion
    uint8_t     titleKey[0x10];           // Encrypted normal-key by key selected by keyyIdx
    uint8_t     reserved0;
    uint64_t    ticketID;               // TicketID
    uint32_t    consoleID;              // ConsoleID
    uint64_t    titleID;                // TitleID
    uint8_t     reserved1[2];
    uint16_t    tikTitleVer;            // Ticket title version
    uint64_t    reserved2;
    uint8_t     licenceType;            // Licence Type
    uint8_t     keyyIdx;                // Index of the common key that used to encrypt this ticket
    uint8_t     reserved3[0x2a];
    uint32_t    accountID;              // eShop Account ID
    uint8_t     reserved4;
    uint8_t     audit;                  // audit ???
    uint8_t     reserved5[0x42];

    // Limits Data
    //uint32_t    limitsIdentifier;       // If the value was 0x4, the ticket is belongs to a demo
    //uint32_t    maxPlayCount;           // Also a Max play count will be set here.
    uint8_t     limitsData[0x40];       // No idea, fill with zeros?

    uint8_t     contentIdx[0xAC];       // Content Index
} TicketData;

typedef struct {
    SignatureData *signatureData;
    TicketData    *tikData;
    void          *pCertChain;          // Pointer points to the head of certification chain
} Ticket_v1;

typedef struct { int fd; void *p, *curp; size_t sz; } tikHandle;

// Forward prototypes
static void SwapBuffer(void *, size_t);
static int ReadBytes(tikHandle *, void *, size_t, int);
static void SkipBytes(tikHandle *, size_t);
static tikHandle *MapTicketFile(const char *);
static void UnmapTicketFile(tikHandle **);
static Ticket_v1 *NewTicketVersion1(void);
static void DestoryTicketVersion1(Ticket_v1 **);
static void PrintRawData(void *, size_t);
static void PrintRSAData(void *, uint32_t);

void ParseSignature(tikHandle *hndl, Ticket_v1 *tikv1)
{
    SignatureData *sigData = NULL;
    uint32_t szData, szPadding; 

    // Ticket->SignatureData->SignatureType
    sigData = malloc(sizeof(sigData));
    ReadBytes(hndl, &sigData->signType, sizeof(uint32_t), 0);
    
    // Ticket->SignatureData->Signature
    szData = getSignDataLength(sigData->signType);
    szPadding = getSignPaddingSize(sigData->signType);
    sigData->signData = calloc(1, szData);
    ReadBytes(hndl, sigData->signData, szData, 0);

    // Skip padding bytes
    SkipBytes(hndl, szPadding);

    tikv1->signatureData = sigData;
}

void ParseTicketData(tikHandle *hndl, Ticket_v1 *tikv1)
{
    TicketData *td = malloc(sizeof(TicketData));
#define _RS(i) ReadBytes(hndl, td->i, sizeof(td->i), 1)
#define _RV(i) ReadBytes(hndl, &td->i, sizeof(td->i),0)
    
    _RS(issuer);
    _RV(pubkey);
    _RV(version);
    _RV(caCrlVersion);
    _RV(signerCrlVersion);
    _RS(titleKey);
    _RV(reserved0);
    _RV(ticketID);
    _RV(consoleID);
    _RV(titleID);
    _RS(reserved1);
    _RV(tikTitleVer);
    _RV(reserved2);
    _RV(licenceType);
    _RV(keyyIdx);
    _RS(reserved3);
    _RV(accountID);
    _RV(reserved4);
    _RV(audit);
    _RS(reserved5);
    _RV(limitsData);
    _RS(contentIdx);

    if(hndl->curp - hndl->p < hndl->sz)
        tikv1->pCertChain = hndl->curp;
    else
        tikv1->pCertChain = NULL;
    tikv1->tikData = td;
#undef _RS
#undef _RV
}

enum { PK_RSA_4096 = 0, PK_RSA_2048, PK_ECC } PubKeyType;
void PrintCertChain(tikHandle *hndl, Ticket_v1 *tikv1)
{
    void *endp = hndl->p + hndl->sz;
    int i = 0;
    uint32_t sigType, sigKeyType;
    uint8_t *sigBody = NULL, sigName[0x40], issuer[0x40], unknown[0x4];
    uint32_t __sigLength, __sigPadding;
#define ReadBytesS(a, b, c, d) do { \
    if(ReadBytes(a, b, c, d)) { \
        if(sigBody) free(sigBody); \
        printf("Error: EOF reached unexpectedly\n"); \
        break; \
    } \
} while(0)

    if(!tikv1->pCertChain) return;

    printf("CERTIFICATION CHAIN:\n");
    while(hndl->curp < endp) {
        printf("  Certificate #%d\n", i++);
        ReadBytesS(hndl, &sigType, sizeof(uint32_t), 0);
        __sigLength = getSignDataLength(sigType);
        __sigPadding = getSignPaddingSize(sigType);
        sigBody = malloc(__sigLength);
        ReadBytesS(hndl, sigBody, __sigLength, 0);
        SkipBytes(hndl, __sigPadding);
        ReadBytesS(hndl, issuer, sizeof(issuer), 1);
        ReadBytesS(hndl, &sigKeyType, sizeof(sigKeyType), 0);
        ReadBytesS(hndl, sigName, sizeof(sigName), 1);
        ReadBytesS(hndl, unknown, sizeof(unknown), 0);

        printf("    Signature Type: %s\n", getSignDesc(sigType));
        printf("    Signature:\n");
        PrintRawData(sigBody, __sigLength);
        printf("    Issuer: %s\n", issuer);
        printf("    Key Type: %s\n", 
            sigKeyType == 0 ? "RSA 4096" : sigKeyType == 1 ? "RSA 2048" : 
            sigKeyType == 2 ? "Elliptic Curve" : "Unknown");
        printf("    Name: %s\n", sigName);
        printf("    Unknown Data:\n");
        PrintRawData(sigName, sizeof(sigName));

        free(sigBody); sigBody = NULL;

        // Parse Public Key
        switch(sigKeyType) {
            case PK_RSA_4096: case PK_RSA_2048: {
                uint32_t szModulus = sigKeyType == PK_RSA_4096 ? 0x200 : 0x100;
                uint32_t pubExponent;
                uint8_t modulus[szModulus];
                ReadBytesS(hndl, modulus, szModulus, 0);
                ReadBytesS(hndl, &pubExponent, sizeof(pubExponent), 0);
                SkipBytes(hndl, 0x34); // fixed padding size

                printf("    RSA %d Modulus:\n", szModulus);
                PrintRawData(modulus, szModulus);
                printf("    Public Exponent: 0x%08X\n", pubExponent);

                break;
            }
            case PK_ECC: {
                uint8_t eccPubKey[0x3c];
                ReadBytesS(hndl, eccPubKey, 0x3c, 0);
                printf("    ECC Public Key:\n");
                PrintRawData(eccPubKey, 0x3c);
                SkipBytes(hndl, 0x3c); // fixed padding size
                break;
            }
        }
    }
}

void PrintTicket(Ticket_v1 *tikv1)
{
    printf("SIGNATURE:\n");
    printf("  Type: %s\n", getSignDesc(tikv1->signatureData->signType));
    printf("  Data:\n");
    PrintRSAData(tikv1->signatureData->signData, tikv1->signatureData->signType);
    putchar('\n');

    TicketData *t = tikv1->tikData;
    printf("TICKET DATA:\n");
    printf("  Issuer: %s\n", t->issuer);
    printf("  ECC Pubkey:\n");
    PrintRawData(t->pubkey, 0x3c);
    printf("  Version: 0x%X\n", t->version);
    printf("  CA CRL Version: 0x%X\n", t->caCrlVersion);
    printf("  Signer CRL Version: 0x%X\n", t->signerCrlVersion);
    printf("  Title Key: %s\n", t->titleKey);
    printf("  Ticket ID: %08llX\n", t->ticketID);
    printf("  Console ID: %08X\n", t->consoleID);
    printf("  Title ID: %08llx\n", t->titleID);
    printf("  Ticket Title Version: 0x%02X\n", t->tikTitleVer);
    printf("  Licence Type: 0x%X\n", t->licenceType);
    printf("  KeyY Index: %d\n", t->keyyIdx);
    printf("  eShop Account ID: %08X\n", t->accountID);
    printf("  Audit: 0x%X\n", t->audit);
    printf("  Limits:\n");
    PrintRawData(t->limitsData, sizeof(t->limitsData));
    printf("  Content Index:\n");
    PrintRawData(t->contentIdx, sizeof(t->contentIdx));
    putchar('\n');
}

int main(int argc, char **argv)
{
    char *fname = NULL;
    tikHandle *tik = NULL;
    Ticket_v1 *tikCtx = NULL;

    if(argc == 2) {
        fname = argv[1];
        if(!(tik = MapTicketFile(fname))) {
            printf("Failed to open Ticket, %s\n", strerror(errno));
            return -1;
        }
        tikCtx = NewTicketVersion1();

        ParseSignature(tik, tikCtx);
        ParseTicketData(tik, tikCtx);

        PrintTicket(tikCtx);
        PrintCertChain(tik, tikCtx);

        DestoryTicketVersion1(&tikCtx);
        UnmapTicketFile(&tik);
    }

    return 0;
}

static int ReadBytes(tikHandle *h, void *p, size_t s, int dontswap)
{
    if(h->curp + s < h->p + h->sz) {
        memcpy(p, h->curp, s);
        if(!dontswap) SwapBuffer(p, s);
        h->curp += s;
        return 0;
    }
    return 1;
}

static void SkipBytes(tikHandle *h, size_t s)
{
    if(h) h->curp += s;
}

static void SwapBuffer(void *pv, size_t n)
{
    char *p = pv; size_t lo, hi;
    for(lo = 0, hi = n - 1; hi > lo; lo++, hi--) {
        p[lo] ^= p[hi];
        p[hi] ^= p[lo];
        p[lo] ^= p[hi];
    }
}

static Ticket_v1 *NewTicketVersion1(void)
{
    Ticket_v1 *t1 = calloc(1, sizeof(Ticket_v1));
    return t1;
}

static void DestoryTicketVersion1(Ticket_v1 **tikv1)
{
    Ticket_v1 *t1 = *tikv1;

    if(t1) {
        if(t1->signatureData) {
            if(t1->signatureData->signData)
                free(t1->signatureData->signData);
            free(t1->signatureData);
        }
        if(t1->tikData)
            free(t1->tikData);
        free(t1);
        *tikv1 = NULL;
    }
}

static tikHandle *MapTicketFile(const char *path)
{
    int fd = -1;
    tikHandle *hndl = NULL;
    struct stat sb;

    if(-1 != (fd = open(path, O_RDONLY))) {
        if(-1 == fstat(fd, &sb))
            goto finished;
        hndl = malloc(sizeof(tikHandle));
        hndl->fd = fd;
        hndl->sz = sb.st_size;
        hndl->p = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
        if(hndl->p == MAP_FAILED)
            UnmapTicketFile(&hndl);
        hndl->curp = hndl->p;
    }
finished:
    return hndl;
}

static void UnmapTicketFile(tikHandle **hndl)
{
    tikHandle *tik = *hndl;

    if(tik) {
        close(tik->fd);
        if(tik->p != MAP_FAILED)
            munmap(tik->p, tik->sz);
        free(tik);
        *hndl = NULL;
    }
}

static void PrintRawData(void *p, size_t sz)
{
    uint8_t *cp = (uint8_t *)p;
    int k = 0, l = 0;

    printf("        ");
    for(int i = 0; i < sz; i++, k++) {
        if(cp[i] == 0xff)
            printf("FF");
        else
            printf("%02X", cp[i]);
        if(k + 1 == 4) {
            l++;
            if(l == 8) {
                putchar('\n'); printf("        "); l = 0;
                k = -1;
                continue;
            }
            putchar(' '); k = -1;
        }
    }
    putchar('\n');
}

static void PrintRSAData(void *sign, uint32_t type)
{
    uint8_t *sp = (uint8_t *)sign;

    for(int i = 0; i < getSignDataLength(type); i += 0x10)
        printf("\t%02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X\n",
            sp[i+0x0], sp[i+0x1], sp[i+0x2], sp[i+0x3],
            sp[i+0x4], sp[i+0x5], sp[i+0x6], sp[i+0x7],
            sp[i+0x8], sp[i+0x9], sp[i+0xA], sp[i+0xB],
            sp[i+0xC], sp[i+0xD], sp[i+0xE], sp[i+0xF]);
}